@@ -7,17 +7,17 @@ This file define classes:
     ParserError,UnifyError: exceptions thrown when parsing or unification fails
       
 """
-import logging, re
+import logging, re, copy, pickle
 from collections import defaultdict
 
 if __name__ == "__main__":
-    from morpher import TurkishPostProcessor
-    from grammar import Grammar,GrammarError,Rule,format_feat
-    from tree import Tree
+    from morpher import TurkishPostProcessor,PostProcessError
+    from grammar import Grammar,GrammarError,Rule,format_feat,format_fparam,Trie
+    from tree import Tree,uid
 else:
-    from .morpher import TurkishPostProcessor
-    from .grammar import Grammar,GrammarError,Rule,format_feat
-    from .tree import Tree
+    from .morpher import TurkishPostProcessor,PostProcessError
+    from .grammar import Grammar,GrammarError,Rule,format_feat,format_fparam,Trie
+    from .tree import Tree,uid
 
 empty_dict = dict()
 empty_set = set()
@@ -71,7 +71,7 @@ class TurkishPreProcessor:
 class DummyPreProcessor:
     """ Dummy pre processor used in Parser """
     def __call__(self,sent):
-        return sent
+        return sent.strip()
 
 class DummyPostProcessor:
     """ Dummy post processor used in Parser """      
@@ -127,7 +127,7 @@ class Parser:
         # for iteration we use set as FIFO, for membership check we use set for O(1) complexity
         for ruleno,rulepos in todo:
             try:
-                symbol = self.rules[ruleno][1][rulepos]
+                symbol = self.rules[ruleno].left[rulepos]
                 for nextno in self.ruledict.get(symbol,set()):
                     nextstate = (nextno,0)
                     if nextstate not in stateset: 
@@ -140,24 +140,26 @@ class Parser:
                     if nextstate not in stateset: 
                         stateset.add(nextstate)
                         todo.append(nextstate)              
-                    symbol = self.rules[ruleno][1][rulepos]
+                    symbol = self.rules[ruleno].left[rulepos]
                 """
             except IndexError: # there is no more symbols in the rule (i.e. A->x.)
                 pass
 
     def format_rules(self):
-         return "\n".join([rule.format() for rule in self.rules])
+        """ return string repr of all rules """
+        return "\n".join(rule.format() for rule in self.rules)
          
     def get_items(self,stateset):
-        """ get string repr of "items" in state set in dotted fomat  e.g "{ S -> NP . VP ; VP -> . V  ; VP -> . V NP }" """
+        """ get string repr of  a "stateset" in state set in dotted fomat  e.g "{ S -> NP . VP ; VP -> . V  ; VP -> . V NP }" """
         slist = ["{"]
         for ruleno,rulepos in stateset:
-            slist.append("{} -> {} . {}".format(self.rules[ruleno][0], " ".join(self.rules[ruleno][1][0:rulepos]), " ".join(self.rules[ruleno][1][rulepos:])))
+            slist.append("{} -> {} . {}".format(self.rules[ruleno].head, " ".join(self.rules[ruleno].left[0:rulepos]), " ".join(self.rules[ruleno].left[rulepos:])))
         slist.append("}")
         return " , ".join(slist)
 
     def get_item(self,ruleno,rulepos):
-        return "{} -> {} . {}".format(self.rules[ruleno][0], " ".join(self.rules[ruleno][1][0:rulepos]), " ".join(self.rules[ruleno][1][rulepos:]))    
+        """ get string repr of a single state item """
+        return "{} -> {} . {}".format(self.rules[ruleno].head, " ".join(self.rules[ruleno].left[0:rulepos]), " ".join(self.rules[ruleno].left[rulepos:]))    
 
     def compile(self):
         """ compile rule list "rules" to a DFA
@@ -177,17 +179,17 @@ class Parser:
         self.ereduce = ereduce
         
         for ruleno,rule in enumerate(rules):
-            ruledict[rule[0]].append(ruleno)
+            ruledict[rule.head].append(ruleno)
 
         # todo: more efficient algorithm for large grammars
-        nullable = {rule[0] for rule in rules if len(rule[1])==0}  # empty productions
+        nullable = {rule.head for rule in rules if len(rule.left)==0}  # empty productions
         flag = bool(nullable)
         while flag:
             flag = False
             for rule in rules:
-                if all(map(lambda x:x in nullable,rule[1])): # all of symbols in RHS is nullable
-                    if rule[0] not in nullable:
-                        nullable.add(rule[0])
+                if all(map(lambda x:x in nullable,rule.left)): # all of symbols in RHS is nullable
+                    if rule.head not in nullable:
+                        nullable.add(rule.head)
                         flag = True
             
         logging.info("nullable=%s", nullable)
@@ -203,7 +205,7 @@ class Parser:
             tempdict = defaultdict(set) # maps symbol to set of next states
             for ruleno,rulepos in states[stateno]:          
                 try:
-                    symbol = rules[ruleno][1][rulepos] # get next symbol after dot (i.e. A->x.By)
+                    symbol = rules[ruleno].left[rulepos] # get next symbol after dot (i.e. A->x.By)
                     tempdict[symbol].add((ruleno,rulepos+1))
                 except IndexError: # there is no more symbols in the rule (i.e. A->x.)
                     pass
@@ -219,7 +221,7 @@ class Parser:
                 dfa[stateno,symbol] = nextstateno
         for idx,stateset in enumerate(states):
             for ruleno,rulepos in stateset:
-                body = rules[ruleno][1]
+                body = rules[ruleno].left
                 if all(map(lambda x:x in nullable,body[rulepos:])): # if all remaining items are nullable
                     if rulepos == 0: # empty or nullable rule
                         ereduce[idx].add((ruleno,rulepos))                   
@@ -228,13 +230,44 @@ class Parser:
         if logging.getLogger().isEnabledFor(logging.DEBUG):
             for stateno,stateset in enumerate(states):
                 logging.debug("%s : %s REDUCE: %s EREDUCE: %s", stateno, self.get_items(stateset), self.get_items(reduce.get(stateno,set())), self.get_items(ereduce.get(stateno,set())))
+
+    def save_grammar(self,fname):
+        with open(fname,"wb") as fout:
+            pickle.dump(self.rules, fout)
+            pickle.dump(self.trie, fout)
+            pickle.dump(self.post_processor.suff_dict, fout)
+            pickle.dump(self.dfa, fout)
+            pickle.dump(self.reduce, fout)
+            pickle.dump(self.ereduce, fout)
+            pickle.dump(self.ruledict, fout)
+
+    def load_grammar(self,fname):
+        with open(fname,"rb") as fin:
+            self.rules = pickle.load(fin)
+            self.trie = pickle.load(fin)
+            self.post_processor.suff_dict = pickle.load(fin)
+            self.dfa = pickle.load(fin)
+            self.reduce = pickle.load(fin)
+            self.ereduce = pickle.load(fin)
+            self.ruledict = pickle.load(fin)
    
-    def load_grammar(self,fname=None,reverse=False,text=None):
+    def parse_grammar(self,fname=None,reverse=False,text=None,defines=None):
         """ loads a grammar file and parse it """
-        self.rules = Grammar.load_grammar(fname,reverse,text)
+        grammar = Grammar.load_grammar(fname,reverse,text,defines)
+        self.rules,self.trie = grammar.rules,grammar.trie
+        #self.post_processor.suff_idxs,self.post_processor.suff_dict_list =  grammar.suff_idxs,grammar.suff_dict_list
+        self.post_processor.suff_dict = grammar.suff_dict
+        
         if logging.getLogger().isEnabledFor(logging.INFO):
             logging.info("rules=%s",self.format_rules())
+            logging.info("dict=%s","\n".join(self.trie.list()))
 
+#all,			// All features are passed through								e.g. V -> PPS
+#only_params,	// Only parameters are passed through							e.g. V -> PPS(+gap) PPS(gap)
+#with_params,	// Parameters and remaining features are passed through			e.g. V -> PPS(+, +gap) 		Note: PPS(+, gap) == PPS
+#without_params	// Except parameters, remaining features are passed through		e.g. V -> PPS(-, gap)		Note: PPS(-, +gap) == PPS(-, gap)
+
+    ### IMPORTANT Currently ?! is ignored in Up Propagation
     def unify_up(dst,param,src):
         """ unification of "src" features into "dst" features, using filtering of "param", if unification fails raises UnifyError
 
@@ -243,43 +276,45 @@ class Parser:
         src is checked against pre-condition dict and then filtered with filter set and unified into dst
 
         """
-        if param is None:
-            keys = src.keys()
+        if hasattr(param,'param_type'):
+            param_type = param.param_type
         else:
-            keys = set()
-            for key in param.keys() & src.keys():
-                if param[key] is None:
-                    keys.add(key)
-                else:
-                    val = param[key]
-                    if val[0] == '*':
-                        val = dst.get(val[1:])
-                    if val:
-                        if val != src[key]:
-                            raise UnifyError("Unify precheck error feat=%s src=%s param=%s" % (key, src[key], val))
+            param_type = None
 
-        newkeys = keys - dst.keys()
-        for key in keys & dst.keys():
-            if  src[key] != dst[key]:
-                if dst[key] == '-':
-                    newkeys.add(key)
-                elif src[key] == '-':
-                    pass
-                elif dst[key] == '+':
-                    newkeys.add(key)
-                elif src[key] == '+':
-                    pass
-                else:
-                    raise UnifyError("Unify error feat=%s src=%s param=%s" % (key, src[key], dst[key]))
+        if param is None:
+            check_keys = []
+            copy_keys = [(key,key) for key in src.keys()]
+        else:
+            check_keys = [(key,val) for key,val in param.items() if val is not None and not val.startswith('*') and key in src]
+            if param_type == '-':
+                copy_keys = [(key,key) for key in src.keys() if key not in param]
+            else:
+                copy_keys = [(key,key if val is None else val[1:]) for key,val in param.items() if (val is None or val.startswith('*')) and key in src]
+                if param_type == '+':
+                    copy_keys.extend((key,key) for key in src.keys() if key not in param)
+            
+        for key,val in check_keys:
+            src_val = src[key]
+            if src_val != val:
+                raise UnifyError("UnifyU precheck error feat=%s src=%s param=%s" % (key, src_val, val))
 
-    
-        if newkeys:
+        new_items = {}
+        for src_key,dst_key in copy_keys:
+            src_val = src[src_key]
+            dst_val = dst.get(dst_key)
+            if dst_val is None:
+                new_items[dst_key] = src_val
+            elif src_val != dst_val:
+                raise UnifyError("UnifyU error feat=%s src=%s dst=%s" % (dst_key, src_val, dst_val))
+
+        if new_items:
             dst = dst.copy()
-            for key in newkeys:
-                dst[key] = src[key]
+            dst.update(new_items)
+
         return dst
 
-    def unify_down(dst,param,src):
+    ### IMPORTANT Currently default copying types +- are ignored in Down Propagation
+    def unify_down(dst,param,src,checklist):
         """ unification of "src" features into "dst" features, using filtering of "param", if unification fails raises UnifyError
 
         if param is None, src is directly unified into dst
@@ -302,27 +337,40 @@ class Parser:
                     if val:
                         pdict[key] = val
 
-        newkeys = pdict.keys() - dst.keys()
         for key in pdict.keys() & dst.keys():
             if pdict[key] != dst[key]:
-                if pdict[key] == '+' and dst[key] != '-':
-                        pass
-                elif dst[key] == '+' and pdict[key] != '-':
-                    newkeys.add(key)
-                else:
-                    raise UnifyError("UnifyD error feat=%s src=%s param=%s" % (key, pdict[key], dst[key]))
+                raise UnifyError("UnifyD error feat=%s src=%s param=%s" % (key, pdict[key], dst[key]))
+
+        newkeys = pdict.keys() - dst.keys()
         if newkeys:
             dst = dst.copy()
             dst.update(pdict)
+
+        for key,val in checklist.items():
+            if val == "?":
+                if key not in dst:
+                    raise UnifyError("UnifyD check error feat not exists %s" % key)
+            elif val == "!":
+                if key in dst:
+                    raise UnifyError("UnifyD check error feat exists %s" % key)
+            elif val.startswith("?"):
+                if key not in dst or val[1:] != dst[key]:
+                    raise UnifyError("UnifyD check error feat not matches %s" % key)
+            elif val.startswith("!"):
+                if key not in dst or val[1:] == dst[key]:
+                    raise UnifyError("UnifyD check error feat matches %s" % key)
+
         return dst
 
- 
     def unify_tree(self,tree):
         """ bottom-up unifies a tree and returns a new tree """
-        rule = self.rules[tree.ruleno]
-        fdict,srcflist = rule.feat, rule.lparam
+
+        rule = tree.rule
+        checklist = rule.checklist
+        fdict = rule.feat
+
         stack = [(fdict,[])]
-        for item,param in zip(tree.left,srcflist):
+        for item,fparam in zip(tree.left,rule.lparam):
             if type(item)==str:
                 for fdict,seq in stack:
                     seq.append(item)
@@ -338,9 +386,9 @@ class Parser:
                             last_error = ue.args[0]
                             continue
                         for subtree in subtrees:
-                            logging.debug("Unify feat=%s param=%s subfeat=%s ", format_feat(fdict), format_feat(param), format_feat(subtree.feat))
-                            try:      
-                                _fdict = Parser.unify_up(fdict,param,subtree.feat)
+                            logging.debug("Unify feat=%s fparam=%s subfeat=%s ", format_feat(fdict), format_fparam(fparam), format_feat(subtree.feat))
+                            try:
+                                _fdict = Parser.unify_up(fdict,fparam,subtree.feat)
                                 logging.debug("Unify Success=%s", format_feat(_fdict))
                                 try:
                                     idx = nkeys.index(_fdict)
@@ -350,8 +398,8 @@ class Parser:
                                     nvals.append([subtree])
                                 #print("nkeys=%s nvals=%s" % (nkeys,nvals))
                             except UnifyError as ue:
-                                logging.debug("Unify Failure dst=%s param=%s src=%s ", format_feat(fdict), format_feat(param), format_feat(subtree.feat))
-                                last_error = "%s super=%s#%d sub=%s#%d" % (ue.args[0], tree.head, tree.ruleno, subtree.head, subtree.ruleno)
+                                logging.debug("Unify Failure dst=%s fparam=%s src=%s ", format_feat(fdict), format_fparam(fparam), format_feat(subtree.feat))
+                                last_error = "%s super=%s#%s sub=%s#%s" % (ue.args[0], tree.head, tree.ruleno, subtree.head, subtree.ruleno)
                     for key,val in zip(nkeys,nvals):
                         nstack.append((key,seq+[val]))
                 stack = nstack
@@ -360,19 +408,22 @@ class Parser:
                     raise UnifyError(last_error)
         ntree = []
         for fdict,seq in stack:
-            ntree.append(Tree(tree.head,tree.ruleno,seq,tree.right,fdict,tree.cost))
+            for key,val in fdict.items():
+                if type(val) == int:
+                    fdict[key] = tree.left[val]
+            ntree.append(Tree(tree.head,tree.rule,tree.ruleno,seq,tree.right,fdict,tree.cost))
         if tree.head=="S'":
             assert len(ntree)==1
             return ntree[0]
         return ntree
 
-    def prune(alts):
-        cost = min(subtree.cost for subtree in alts)
-        if cost<0:
-            alts = [subtree for subtree in alts if subtree.cost==cost]
-            for subtree in alts:
-                subtree.cost = 0
-        return alts
+    #def prune(alts):
+    #    cost = min(subtree.cost for subtree in alts)
+    #    if cost<0:
+    #        alts = [subtree for subtree in alts if subtree.cost==cost]
+    #        for subtree in alts:
+    #            subtree.cost = 0
+    #    return alts
 
     def make_trans_tree(self,symbol,feat,fparam):
         """ generate a tree for dst-only non-terminal tree """
@@ -380,54 +431,112 @@ class Parser:
         for ruleno in self.ruledict[symbol]:
             rule = self.rules[ruleno]
             try:
-                fdict = Parser.unify_down(rule.feat,fparam,feat)
-                logging.debug("make_trans_tree: %s unifyd(%s,%s,%s)->%s", symbol, format_feat(feat), format_feat(fparam,'()'), format_feat(rule.feat), format_feat(fdict))
-                sub = []
+                #logging.debug("make_trans_tree1: %s unifyd(%s,%s,%s)", symbol, format_feat(feat), format_feat(fparam,'()'), format_feat(rule.feat))
+                fdict = Parser.unify_down(rule.feat,fparam,feat,rule.checklist)
+                logging.debug("make_trans_tree: %s unifyd(%s,%s,%s)->%s", symbol, format_feat(feat), format_fparam(fparam), format_feat(rule.feat), format_feat(fdict))
+                trans = []
                 for item,param in zip(rule.right,rule.rparam):
-                    assert type(item) != int, "%s#%d(%s) %s:%s" % (symbol,ruleno,fparam,item,param)
-                    if param is False: # Terminal
-                        sub.append(item)
+                    assert type(item) == str, "%s#%d(%s) %s:%s" % (symbol,ruleno,fparam,item,param)
+                    if item[0]=='*': # Reference
+                        key = item[1:]
+                        try:
+                            val = fdict[key]
+                        except KeyError:
+                            raise UnifyError("Translate Ref to feature %s not found in %s" % (key,rule))
+                        if type(val)==str:
+                            if val[0].isupper():
+                                trans.append( self.make_trans_tree(val,fdict,param) )
+                            else:
+                                trans.append(val)
+                        else:
+                            assert param is not False
+                            trans.append( self.trans_alts(val,fdict,param) )
+                    elif param is False: # Terminal
+                        trans.append(item)
                     else: # NonTerminal
-                        sub.append(Parser.prune( self.make_trans_tree(item,fdict,param) ))
-                ntree.append(Tree(symbol,ruleno,[],sub,fdict,rule.cost)) 
+                        trans.append( self.make_trans_tree(item,fdict,param) )
+
+                ntree.append(Tree(symbol,rule,ruleno,[],trans,fdict,rule.cost))
+                if rule.cut:
+                    break
             except UnifyError as ue:
                 last_error = "%s %s#%d" % (ue.args[0], symbol, ruleno)
-                logging.debug("make_trans_tree: %s unifyd(%s,%s,%s)->Error", symbol, format_feat(feat), format_feat(fparam,'()'), format_feat(rule.feat))
+                logging.debug("make_trans_tree: %s unifyd(%s,%s,%s)->Error", symbol, format_feat(feat), format_fparam(fparam), format_feat(rule.feat))
         if not ntree:
+            if not self.ruledict[symbol]:
+                raise ParseError("No production rule defined for symbol: %s" % symbol)
             raise UnifyError(last_error)
         return ntree
 
+    def trans_alts(self,subtrees,fdict,param):
+        assert type(subtrees) == list
+        alts = []
+        for alt in subtrees:
+            try:
+                alts.append( self.trans_tree(alt,fdict,param) )
+                if alt.rule.cut:
+                    break
+            except UnifyError as ue:
+                last_error = "%s %s" % (ue.args[0], alt.rule)
+        if not alts:
+            raise UnifyError(last_error)
+        return alts
+
     def trans_tree(self,tree,feat=empty_dict,fparam=None):
         """ translates and unifies translation(right) part of a parse tree
         returns a modified version of the node "tree" """
 
         assert type(tree)==Tree
-        logging.debug("trans_tree: %s unify(%s,%s,%s)->", tree.head, format_feat(tree.feat), format_feat(fparam,'()'), format_feat(feat))                       
-        fdict = Parser.unify_down(tree.feat,fparam,feat)
-        logging.debug("trans_tree: ->%s", format_feat(fdict))                       
-        rule = self.rules[tree.ruleno]
+        tree = copy.copy(tree) # MD 12.04.2018
+        rule = tree.rule
+
+        logging.debug("trans_tree: %s unify(%s,%s,%s)->", tree.head, format_feat(tree.feat), format_fparam(fparam), format_feat(feat)) 
+        fdict = Parser.unify_down(tree.feat, fparam,feat, rule.checklist)
+        logging.debug("trans_tree: ->%s", format_feat(fdict)) 
+        
         trans = []
         for item,param in zip(rule.right,rule.rparam):
-            if param is False: # Terminal
+            if type(item)==int: # Matched (Left&Right) NT
+                assert param is not False
+                trans.append( self.trans_alts(tree.left[item],fdict,param) )
+            elif item[0]=='*': # Reference
+                assert param is not False
+                key = item[1:]
+                try:
+                    val = fdict[key]
+                except KeyError:
+                    raise UnifyError("Translate Ref to feature %s not found in %s" % (key,rule))
+                if type(val)==str:
+                    if val[0].isupper():
+                        trans.append( self.make_trans_tree(val,fdict,param) )
+                    else:
+                        trans.append(val)
+                else:
+                    trans.append( self.trans_alts(val,fdict,param) )
+            elif param is False: # Terminal
                 trans.append(item)
-            elif type(item)==str: # Unmatched (Right-Only) NT
+            else: # Unmatched (Right-Only) NT
                 assert item[0].isupper()
-                trans.append(Parser.prune( self.make_trans_tree(item,fdict,param) )) 
-            else: # Matched (Left&Right) NT
-                assert type(item)==int
-                trans.append(Parser.prune( [self.trans_tree(alt,fdict,param) for alt in tree.left[item]] ))           
+                trans.append( self.make_trans_tree(item,fdict,param) ) 
+        
         tree.right = trans
         return tree
 
     def format_edge_item(alts):
         """ internal: format an edge item to str (used internally for logging/debugging) """
-        return " ".join(["{2}({0},{1};{3},{4})".format(*alt) for alt in alts[1:]])
+        return " ".join(["{2}({0},{1};{3},{4})".format(*alt) if type(alt)==tuple else " ".join(alt)
+                         for alt in alts[1:]])
       
     def format_edge(self,edge):
         """ internal: format an edge into str (used internally for logging/debugging) """
         if edge not in self.edges:
-            return "{2}({0},{1};{3},{4}) -> None".format(*edge)
-        return  "{2}({0},{1};{3},{4}) -> ".format(*edge) + " | ".join( [ " ".join(["{2}({0},{1};{3},{4})".format(*alt) for alt in alts[1:]]) for alts in self.edges[edge]] )       
+            if type(edge)==tuple:
+                return "{2}({0},{1};{3},{4}) -> None".format(*edge)
+            else:
+                return " ".join(edge)
+        return  "{2}({0},{1};{3},{4}) -> ".format(*edge) + " | ".join( [ " ".join(
+            ["{2}({0},{1};{3},{4})".format(*alt) if type(alt)==tuple else alt
+             for alt in alts[1:]]) for alts in self.edges[edge]] )
     
     def print_parse_tables(self):
         """ internal: print parse tables after parse """
@@ -443,6 +552,7 @@ class Parser:
     def make_tree(self):
         self.tree = Tree(
             head = "S'",
+            rule = self.rules[0],
             ruleno = 0,
             left = [self.make_tree_int(self.top_edge)]
         )
@@ -451,15 +561,28 @@ class Parser:
     def make_tree_int(self,edge):
         """ generates a tree (which is a recursive list of lists) from edges """
         if edge not in self.edges:
-            return edge[2]
+            if type(edge)==tuple: #(state1,pos1,term,state2,pos2)
+                return edge[2]
+            else: # list of non-terminals
+                return edge;
         alt = []
         for alt_edge in self.edges[edge]: 
-            rule = self.rules[alt_edge[0]] 
+            ruleno = alt_edge[0]
+            if type(ruleno)==int:
+                rule = self.rules[ruleno]
+            else:
+                rule = ruleno
+                ruleno = None
+
+            assert type(ruleno)==int or ruleno is None, "ruleno=%s" % ruleno
+            assert type(rule) == Rule
+
             alt.append( Tree(
                 head = edge[2],
-                ruleno = alt_edge[0],
+                rule = rule,
+                ruleno = ruleno, 
                 left = [self.make_tree_int(sub_edge) for sub_edge in alt_edge[1:]],
-                right = rule.right,
+                right = rule.right,#.copy(),
                 feat = rule.feat,
                 cost = rule.cost
             ) )
@@ -501,21 +624,26 @@ class Parser:
         self.instr = instr
         self.top_edge = (0,0,"S",inlen-1,fstate)
 
-        active = {0}
-        rset = set()
+        act_states = [set() for i in range(inlen)] # active set of states for each position
+        act_edges  = [set() for i in range(inlen)] # active set of edges for each position
+
+        act_states[0].add(0); # add initial state to initial position
+
         
         for pos in range(inlen):
-            token = instr[pos]         
+            token = instr[pos]
+             
+            rlist = list(act_edges[pos])
+            active = act_states[pos]
 
-            logging.debug("eset=%s rset=%s", active, rset)
-
-            rlist = list(rset)
+            logging.debug("act_states=%s act_edges=%s", active, rlist)
 
             for edge in rlist: # for each work item (start_position, start_state, edge_symbol, end_position, end_state)
                 spos,sstate,esymbol,epos,estate = edge
                 logging.debug("Checking Work Item: %s  All: %s", edge, rlist)
                 for ruleno,rulepos in reduce.get(estate,set()): # find reducible items for end_state
-                    head,body = self.rules[ruleno][0:2]
+                    head = self.rules[ruleno].head
+                    body = self.rules[ruleno].left
                     logging.debug("Reducing %s ", self.get_item(ruleno,rulepos))
                     ptree = [edge]
                         
@@ -536,7 +664,7 @@ class Parser:
                                     nstack.append((xpos,xstate,[(xpos,xstate,symbol,ppos,pstate)]+ptree))
                                     
                             stack = nstack
-                    for ppos,pstate,ptree in stack:                      
+                    for ppos,pstate,ptree in stack:
                         nstate = self.dfa.get((pstate,head),-1)
                         logging.debug("REDUCE %s , %s -> %s", pstate, head, nstate)
                         if nstate != -1:
@@ -550,10 +678,12 @@ class Parser:
                             edges[nedge].append(ptree)
 
             actlist = list(active)
+
             for state in actlist:
                 for ruleno,rulepos in ereduce.get(state,set()):
                     logging.debug("e-Reducing %s", self.get_item(ruleno,rulepos))
-                    head,body = self.rules[ruleno][0:2]
+                    head = self.rules[ruleno].head
+                    body = self.rules[ruleno].left
                     ptree = [ruleno]
                     estate = state
                     for symbol in body[rulepos:]:
@@ -576,14 +706,17 @@ class Parser:
                 if fstate in active:
                     logging.info("Parse successful")
                 else:
-                    logging.error("Parse not possible")
-                    raise ParseError("Parse not possible")
+                    while not act_states[pos]:
+                        pos -= 1
+                    error = "Parse is not possible at position %s: %s [* %s *] %s" % (pos, " ".join(instr[0:pos]),instr[pos]," ".join(instr[pos+1:-1]))
+                    logging.error(error)
+                    raise ParseError(error)
             else:
                 if not active:
-                    logging.error("not active, %s",active)
-                    raise ParseError("Cannot shift %s<< %s" % (" ".join(instr[0:pos])," ".join(instr[pos:])))
-                rset = set()
-                nactive = set()
+                    continue
+                    #logging.error("not active, %s",active)
+                    #raise ParseError("Cannot shift %s<< %s" % (" ".join(instr[0:pos])," ".join(instr[pos:])))
+
                 for state in active:
                     nstate = dfa.get((state,token),-1)
                     #print(state,",",token,"->",nstate)
@@ -592,9 +725,28 @@ class Parser:
                         nodes[pos+1,nstate,token].add((pos,state))
                         logging.debug("shift %s = %s", (pos+1,nstate,token,pos,state),token)
                     
-                        rset.add((pos,state,token,pos+1,nstate))
-                        nactive.add(nstate)
-                active = nactive
+                        act_edges[pos+1].add((pos,state,token,pos+1,nstate))
+                        act_states[pos+1].add(nstate)
+
+                items = self.trie.search(instr[pos:])
+                logging.debug("Shift pos: %d items: %s", pos, items)
+                #items.append((1,(token,[],[]))
+                for input_len,rule in items:
+                    token = rule.head
+                    nextpos = pos + input_len
+                    for state in active:
+                        nstate = dfa.get((state,token),-1)
+                        if nstate != -1:                           
+                            logging.debug("add nodes[%s] = %s",(nextpos,nstate,token),(pos,state)) 
+                            nodes[nextpos,nstate,token].add((pos,state))
+                            nedge = (pos,state,token,nextpos,nstate)
+                            logging.debug("shift %s = %s", nedge, token)
+                    
+                            act_edges[nextpos].add(nedge)
+                            act_states[nextpos].add(nstate)
+
+                            edges[nedge].append([rule]+rule.left)
+                            #logging.debug("appending edge %s to %s", Parser.format_edge_item(ptree), self.format_edge(nedge))
 
     def trans_sent(self,sent):
         """ translates a sentence, returns a list of possible translations or an error """
@@ -604,86 +756,15 @@ class Parser:
             tree = self.make_tree()
             tree2 = self.unify_tree(tree)
             tree3 = self.trans_tree(tree2)
-            return [(self.post_processor(trans),cost) for trans,cost in tree3.enumx()]
+
+            result = [(self.post_processor(trans),cost) for trans,cost in tree3.enumx()]
+            result.sort(key=lambda item:item[1])
+            return result
+            #return list(tree3.enumx())
         except ParseError as pe:
-            return str(pe)
+            return "ParseError: "+str(pe)
         except UnifyError as ue:
-            return str(ue)
-        
-    
-    def trans_file(self,infile,outfile):
-        """ parses all sentences in infile. Each line should be in the form: InputSentence [ "@" ExpectedTranslation ]
-        input and corresponding translations are written to output file. The file is appended by statistics (InputCount,TranslatedCount,MatchedCount)
-        """
-        input_cnt = 0
-        trans_cnt = 0
-        match_cnt = 0
-        experr_cnt = 0
-        pre_processor = DummyPreProcessor()
-        with open(infile, 'r') as fin, open(outfile, 'w') as fout:
-            for line in fin:
-                line = line.strip()
-                if not line:
-                    continue
-                input_cnt += 1
-                parts = line.split('@')
-                if len(parts) == 2: # input file contains the expected translation
-                    if self.reverse:
-                        trans,sent = parts
-                    else:
-                        sent,trans = parts
-                    trans = [pre_processor(tsent) for tsent in trans.split('|')]
-                else:
-                    if sent.reverse:
-                        raise ParseError("Input sentence doesn't contain translation", line)
-                    sent = parts[0]
-                    trans = []
-
-                sent = sent.strip()
-                if sent == '*': # reverse translation for expected parse error
-                    continue
-                        
-                print("input:"," @ ".join([sent.strip()," | ".join(trans)]),file=fout)
-                print("output:",file=fout,end=" ")
-                trans_list = self.trans_sent(sent)
-                #print(sent, trans_list)
-                if type(trans_list)==str: # an error occured
-                    if '*' in trans:
-                        experr_cnt += 1
-                        print("EXPECTED ERROR",file=fout)
-                    else:
-                        print("ERROR",file=fout)
-                    print(trans_list,file=fout)
-                else:
-                    trans_list, trans_cost = zip(*trans_list)
-                    trans_cnt += 1
-                    if any(tsent in trans_list for tsent in trans):
-                        print("OK",file=fout)
-                        match_cnt += 1  
-                    else:
-                        print("NOK",file=fout)
-                    for idx, alt in enumerate(trans_list):
-                        print(idx+1,alt,trans_cost[idx],file=fout)
-                print(file=fout)
-            print("input={}, translated={}, matched={} exp_err={}".format(input_cnt,trans_cnt,match_cnt,experr_cnt),file=fout)
-        return (input_cnt,trans_cnt,match_cnt,experr_cnt)
-            
-
-def main():
-    cases = [ # dst,param,src,exp_result(None if exception)
-        ({},{"ng":"1"},{"det":"1"},{"ng":"1"}),
-        ({},None,{"det":"1"},{"det":"1"}),
-        ({"det":"1","numb":"plur"},None,{"det":"1","pers":"1"},{"det":"1","numb":"plur","pers":"1"}),
-        ({"det":"0","numb":"plur"},None,{"det":"1","pers":"1"},None),
-        ({"det":"0","numb":"plur"},{"pers":None},{"det":"1","pers":"1"},{"det":"0","numb":"plur","pers":"1"}),
-    ]
-    for dst,param,src,exp in cases:
-        try:
-            result = Parser.unify_down(dst,param,src)
-        except UnifyError:
-            result = None
-        print(result,exp,"OK" if result==exp else "NOK")
-    
-if __name__ == "__main__":
-    # execute only if run as a script
-    main()
+            return "UnifyError: "+str(ue)
+        except PostProcessError as ppe:
+            return "PostProcessError: "+str(ppe)
+        
\ No newline at end of file
